<?php
/**
 * DNSthingy class.
 *
 * @category   apps
 * @package    dnsthingy
 * @subpackage libraries
 * @author     DNSthingy <support@dnsthingy.com>
 * @copyright  2015-2017 DNSthingy
 * @license    All rights reserved
 * @link       https://www.dnsthingy.com
 */

///////////////////////////////////////////////////////////////////////////////
// N A M E S P A C E
///////////////////////////////////////////////////////////////////////////////

namespace clearos\apps\gateway_management;

///////////////////////////////////////////////////////////////////////////////
// B O O T S T R A P
///////////////////////////////////////////////////////////////////////////////

$bootstrap = getenv('CLEAROS_BOOTSTRAP') ? getenv('CLEAROS_BOOTSTRAP') : '/usr/clearos/framework/shared';
require_once $bootstrap . '/bootstrap.php';

///////////////////////////////////////////////////////////////////////////////
// T R A N S L A T I O N S
///////////////////////////////////////////////////////////////////////////////

clearos_load_language('base');
clearos_load_language('gateway_management');


///////////////////////////////////////////////////////////////////////////////
// D E P E N D E N C I E S
///////////////////////////////////////////////////////////////////////////////

// Classes
//--------

use \clearos\apps\base\Configuration_File;
use \clearos\apps\base\Daemon;
use \clearos\apps\base\Engine;
use \clearos\apps\base\File;
use \clearos\apps\base\Folder;
use \clearos\apps\base\Shell;
use \clearos\apps\clearcenter\Rest;
use \clearos\apps\dhcp\Dnsmasq;
use \clearos\apps\network\Hosts;
use \clearos\apps\network\Iface;
use \clearos\apps\network\Iface_Manager;
use \clearos\apps\network\Network;
use \clearos\apps\network\Network_Utils;
use \clearos\apps\network\Resolver;
use \clearos\apps\tasks\Cron;
use \clearos\apps\marketplace\Marketplace as Marketplace;

clearos_load_library('base/Configuration_File');
clearos_load_library('base/Daemon');
clearos_load_library('base/Engine');
clearos_load_library('base/File');
clearos_load_library('base/Folder');
clearos_load_library('base/Shell');
clearos_load_library('clearcenter/Rest');
clearos_load_library('dhcp/Dnsmasq');
clearos_load_library('network/Hosts');
clearos_load_library('network/Iface');
clearos_load_library('network/Iface_Manager');
clearos_load_library('network/Network');
clearos_load_library('network/Network_Utils');
clearos_load_library('network/Resolver');
clearos_load_library('tasks/Cron');
clearos_load_library('marketplace/Marketplace');

// Exceptions
//-----------
use \clearos\apps\base\Engine_Exception;

clearos_load_library('base/Engine_Exception');


///////////////////////////////////////////////////////////////////////////////
// C L A S S
///////////////////////////////////////////////////////////////////////////////

/**
 * DNSthingy interface class.
 *
 * @category   apps
 * @package    gateway_management
 * @subpackage libraries
 * @author     DNSthingy
 * @copyright  2014 DNSthingy
 * @license    All rights reserved
 * @link       http://www.dnsthingy.com/
 */

class DNSthingy extends Daemon
{
    ///////////////////////////////////////////////////////////////////////////////
    // C O N S T A N T S
    ///////////////////////////////////////////////////////////////////////////////

    // Misc.
    const API_URL             = 'https://dashboard.contentfilter.net/';

    // Config
    const BLOCKPAGE_PORT      = '6373';
    const DNSTHINGY_DIR       = '/etc/dnsthingy';
    const DT_APP              = '/etc/dnsthingy/app.conf';
    const DT_MGR_CONFIG       = '/etc/dnsthingy/dnsthingymgr.conf';
    const DT_IPE_CONFIG       = '/etc/dnsthingy/dnsthingyipe.conf';
    const DNSMASQ_CONFIG      = '/etc/dnsmasq.d/app-dnsthingy.conf';
    const DNSMASQ_MAIN_CONFIG = '/etc/dnsmasq.conf';
    const FIREWALL_RULES      = '/etc/clearos/firewall.d/10-dnsthingy';
    const CHECK_ALIVE         = '/usr/local/sbin/dnsthingycheck.sh';

    // Key paths
    const PATH_PK_FILE        = '/etc/dnsthingy/signature.pk';
    const PATH_SK_FILE        = '/etc/dnsthingy/signature.sk';
    const PEM_CERT_FILE       = '/etc/dnsthingy/dt_cert.pem';
    const PEM_DH_FILE         = '/etc/dnsthingy/dh.pem';
    const PEM_KEY_FILE        = '/etc/dnsthingy/privkey.pem';

    ///////////////////////////////////////////////////////////////////////////////
    // V A R I A B L E S
    ///////////////////////////////////////////////////////////////////////////////

    protected $config = NULL;
    protected $is_loaded = array();
    protected $is_gateway = NULL;
    protected $issues = array();
    protected $first_if = NULL;
    protected $grep_path = NULL;
    protected $ip_path = NULL;
    protected $ping_path = NULL;
    protected $internal_dns = NULL;

    ///////////////////////////////////////////////////////////////////////////////
    // M E T H O D S
    ///////////////////////////////////////////////////////////////////////////////

    /**
     * DNSthingy constructor.
     */
    public function __construct()
    {
        clearos_profile(__METHOD__, __LINE__);

        parent::__construct('dnsthingymgr');

        // check if this is a gateway
        $network = new Network();
        $mode = $network->get_mode();
        if ($mode == 'gateway')
            $this->is_gateway = TRUE;
        else
            $this->is_gateway = FALSE;
    }

    /**
     * Setup all config files
     *
     * @return void
     */
    public function setup()
    {
        clearos_profile(__METHOD__, __LINE__);

        $config = $this->config('dt_mgr');

        // Versions with redirect-local set could not be disabled, so we will
        //  want to enable this new version as well.
        if (isset($config['redirect-local'])) {
            $this->_set_parameter('dt_app', 'enabled');
        }

        // Set up
        $this->_create_keys();
        $this->_setup_dt_config();
        $this->_setup_ipe_config();

        if ($this->issues)
            return FALSE;
        else
            return TRUE;
    }

    /**
     * Verify settings are valid.
     *
     * @return array error messages
     */
    public function verify_setup()
    {
        clearos_profile(__METHOD__, __LINE__);

        $this->update_config();

        // if needed, restart daemons to pick up any config changes
        if ($this->is_enabled() && count($this->get_issues())) {
            $this->restart_daemons();
        }

        // check firewall settings
        $issue = $this->update_firewall();
        $this->add_issue($issue);

        $this->update_daemons();

        $this->update_checkalive();
    }

    /**
     * Add an issue to the list.
     *
     * @param string $issue Issue description
     *
     * @return void
     */
    public function add_issue($issue)
    {
        clearos_profile(__METHOD__, __LINE__);

        if ($issue) {
            $this->issues[] = $issue;
        }
    }

    /**
     * Returns an array of issues that need to be resolved before automatic setup can be completed
     *
     * @return array issues
     */
    public function get_issues()
    {
        clearos_profile(__METHOD__, __LINE__);

        return $this->issues;
    }

    /**
     * Reset list of issues. Used to clear false reports after enabling or disabling.
     *
     * @return void
     */
    public function reset_issues()
    {
        clearos_profile(__METHOD__, __LINE__);

        $this->issues = array();
    }

    /**
     * Get enabled status.
     *
     * @return bool
     */
    public function is_enabled()
    {
        clearos_profile(__METHOD__, __LINE__);

        $config = $this->config('dt_app');
        return isset($config['enabled']);

    }

/**
     * Set enabled status.
     *
     * @param bool $enabled Enabled setting
     *
     * @return bool updated
     */
    public function set_enabled($enabled)
    {
        clearos_profile(__METHOD__, __LINE__);

        if ($this->is_enabled() === $enabled)
            return FALSE;

        if ($enabled) {
            $this->enable();
        } else {
            $this->disable();
        }

        $this->reset_issues();

        return TRUE;
    }

    /**
     * Get HTTP listen address.
     *
     * @return string
     */
    public function get_http_address()
    {
        clearos_profile(__METHOD__, __LINE__);

        $config = $this->config('dt_mgr');
        if (isset($config['http-listen-address'])) {
            return $config['http-listen-address'];
        } else {
            return NULL;
        }

    }

    /**
     * Set http listen address.
     *
     * @param bool $ip IP address
     *
     * @return bool updated
     */
    public function set_http_address($ip)
    {
        clearos_profile(__METHOD__, __LINE__);

        if ($this->get_http_address() === $ip)
            return FALSE;

        $this->_set_parameter('dt_mgr', 'http-listen-address', $ip);

        return TRUE;
    }

    /**
     * Get marketplace info.
     *
     * @return string
     */
    public function get_marketplace_info()
    {
        clearos_profile(__METHOD__, __LINE__);

        $marketplace = new Marketplace();
        $data = $marketplace->get_sdn_info();

        return $data;
    }

    /**
     * Register.
     *
     * @return array
     */
    public function register()
    {
        clearos_profile(__METHOD__, __LINE__);

        // get the post data
        $marketplace = $this->get_marketplace_info();
        $info = json_decode($marketplace);

        $post = array(
            'email' => $info->email,
            'name' => $info->username,
            'billing_id' => (string) $info->username,
            'billing_service' => 'clearos',
            'network_name' => $info->device_name,
            'box_id' => $this->get_box_id(),
            'device_billing_id' => (string) $info->device_id,
        );

        $post = json_encode($post);

        $url = self::API_URL . 'api/v2/partner/clients';

        $ch = curl_init();
        curl_setopt($ch, CURLOPT_URL, $url);

        curl_setopt($ch, CURLOPT_HTTPHEADER, array("Content-type: application/json"));
        curl_setopt($ch, CURLOPT_POSTFIELDS, $post);

        curl_setopt($ch, CURLOPT_RETURNTRANSFER, 1);
        curl_setopt($ch, CURLOPT_HEADER, 1);
        curl_setopt($ch, CURLOPT_TIMEOUT, 60);
        curl_setopt($ch, CURLOPT_SSL_VERIFYHOST, 0);
        curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, FALSE);
        $response = curl_exec($ch);
        $h_size = curl_getinfo($ch, CURLINFO_HEADER_SIZE);
        curl_close($ch);

        list($headers, $body) = explode("\r\n\r\n", $response, 2);
        preg_match('/^HTTP\/.*$/m', $headers, $status);

        return array(
            'header' => $status[0],
            'body' => $body,
        );
    }

    /**
     * Enable.
     *
     * @return bool
     */
    public function enable()
    {
        clearos_profile(__METHOD__, __LINE__);

        // 1. Set enabled.
        $this->_set_parameter('dt_app', 'enabled');

        // 2. Update all conf files. Must be done before restarting daemons.
        $this->update_config();

        // 3. Add firewall rules.
        $this->update_firewall();

        // 4. Restart the daemons. Filtering should start working.
        $this->restart_daemons();

        // 5. Configure daemon startup scripts.
        $this->update_daemons();

        // 6. Add checkalive cron.
        $this->update_checkalive();
    }

    /**
     * Disable.
     *
     * @return bool
     */
    public function disable()
    {
        clearos_profile(__METHOD__, __LINE__);

        // 1. Remove enabled.
        $this->_set_parameter('dt_app', 'enabled', NULL, TRUE);

        // 2. Remove checkalive cron.
        $this->update_checkalive();

        // 3. Update config.
        $this->update_config();

        // 4. Restart dnsmasq.
        $this->_daemon_cmd('dnsmasq', 'restart');

        // 5. Remove firewall rules.
        $this->update_firewall();

        // 6. Handle the daemons.
        $this->update_daemons();
    }

    /**
     * Update checkalive.
     *
     * @return void
     */
    public function update_checkalive()
    {
        clearos_profile(__METHOD__, __LINE__);

        $cron = new Cron();

        $enabled = $this->is_enabled();

        // Make sure checkalive script is there
        $file = new File(self::CHECK_ALIVE);
        if (!$file->exists()) {
            $this->add_issue(lang('gateway_management_check_alive_not_found'));
            $enabled = FALSE;
        } elseif ($file->get_permissions() != '0755') {
            $this->add_issue(lang('gateway_management_check_alive_not_executable'));
            $enabled = FALSE;
        }

        if ($enabled) {
            // Enabled, so make sure checkalive script is active.
            if (!$cron->exists_configlet('app-dnsthingy')) {
                $check_alive = self::CHECK_ALIVE;
                $configlet  = "SHELL=/bin/bash\n";
                $configlet .= "PATH=/sbin:/bin:/usr/sbin:/usr/bin:/usr/local/sbin\n";
                $configlet .= "* * * * * root $check_alive >/dev/null 2>&1";
                $cron->add_configlet('app-dnsthingy', $configlet);
                $this->add_issue(lang('gateway_management_check_cron_not_found'));
            }

        } else {
            // Disabled, so make sure checkalive does not happen.
            if ($cron->exists_configlet('app-dnsthingy')) {
                $cron->delete_configlet('app-dnsthingy');
            }

            // Kill checkalive script if it is running.
            //   This will prevent false alerts since it can sleep for up to 59 seconds.
            $shell = new Shell();
            try {
                $shell->execute('/usr/bin/killall', 'dnsthingycheck.sh');
            } catch (Engine_Exception $e) {
                // not running, so everything's fine
            }
        }
    }

    /**
     * Update config.
     *
     * @return void
     */
    public function update_config()
    {
        clearos_profile(__METHOD__, __LINE__);

        $configs = $this->standard_config();

        if ($this->is_enabled()) {

            // add defaults
            $configs['dnsmasq']['add-subnet'] = '32,128';
            $configs['dnsmasq']['no-poll'] = NULL;
            $configs['dnsmasq']['no-resolv'] = NULL;
            $configs['dnsmasq']['server'] = '127.0.0.1#5553';
            $configs['dnsmasq_main']['cache-size'] = '0';

            // remove invalid options
            $configs['dnsmasq']['bind-interfaces'] = FALSE;
            $configs['dnsmasq']['listen-address'] = FALSE;
            $configs['dnsmasq_main']['bogus-priv'] = FALSE;

            // validate http-listen-address
            $http_address = $this->get_http_address();
            $result = $this->validate_http_address($http_address);
            if ($result) {
                $this->add_issue(lang('gateway_management_blockpage_ip_autoupdated'));
                if ($this->is_enabled()) {
                    $http_address = $this->suggest_http_address();
                    $this->_set_parameter('dt_mgr', 'http-listen-address', $http_address);
                }
            }

        } else {

            $file = new File(self::DNSMASQ_CONFIG);
            if ($file->exists()) {
                $file->delete();
            }

            $this->_reset_parameter('dnsmasq_main', 'cache-size');
            $this->_reset_parameter('dnsmasq_main', 'bogus-priv');
        }

        // check all the config values
        foreach ($configs as $file => $items) {
            foreach ($items as $k => $v) {
                $delete = ($v === FALSE ? TRUE : FALSE);
                $was_set = $this->_verify_parameter($file, $k, $v, $delete);
                if (!$was_set) {
                    $issue = "'$k'".lang('gateway_management_cfg_err')."'$file'.";
                    $this->add_issue($issue);
                }
            }
        }
    }

    /**
     * Standard config options.
     *
     * @return array
     */
    public function standard_config() {
        $lans = $this->get_lan_interfaces();
        sort($lans);
        $lans = implode(',', $lans);

        // needed config values
        $configs = array(
            'dt_mgr' => array(
                'daemon' => NULL,
                'dhcplease-file' => '/var/lib/dnsmasq/dnsmasq.leases',
                'dns-listen-address' => '127.0.0.1@5553',
                'edns0-accept' => NULL,
                'enable-tuntap' => NULL,
                'http-listen-port' => self::BLOCKPAGE_PORT,
                'ipe-server-address' => '127.0.0.1@38473',
                'lan-iface' => $lans,
                'lock-file' => '/var/lock/dnsthingymgr.lock',
                'pid-file' => '/var/run/dnsthingymgr.pid',
                'resolv-file' => '/etc/resolv-peerdns.conf',
                'signature-pk-file' => self::PATH_PK_FILE,
                'signature-sk-file' => self::PATH_SK_FILE,
                'ssl-ca-file' => '/etc/dnsthingy/cert.pem',
                // 'log-level' => '6',
                // 'log-file' => '/var/log/dnsthingymgr.log',
                // mybox-redirect - autoset below

                // delete
                'dtts-dumpfile' => FALSE,
                'dtts-minage' => FALSE,
            ),
            'dt_ipe' => array(
                'capture-iface' => $lans,
                'daemon' => NULL,
                'listen-address' => '127.0.0.1@38473',
                'lock-file' => '/var/lock/dnsthingyipe.lock',
                'pid-file' => '/var/run/dnsthingyipe.pid',
                'signature-pk-file' => '/etc/dnsthingy/signature.pk',
                'ssl-ca-file' => '/etc/dnsthingy/cert.pem',
                'ssl-cert-file' => '/etc/dnsthingy/dt_cert.pem',
                'ssl-dh-file' => '/etc/dnsthingy/dh.pem',
                'ssl-pk-file' => '/etc/dnsthingy/privkey.pem',
                'ssl-pk-pwd' => 'dnsthingyIPE',
                // 'log-level' => '6',
                // 'log-file' => '/var/log/dnsthingyipe.log',
            ),
        );

        // get internal interface
        $iface = new Iface($this->get_lan_interface());
        $lan_ip = $iface->get_live_ip();

        // set up mybox redirect
        if ($lan_ip) {
            // TODO: Fix below command. It erases any user created host entries
            //       for the lan ip. Leave disabled until fixed.
            // setup hosts entries
            // $host_result = $this->_setup_host('mybox.management', $lan_ip);
            // if ($host_result) {
            //     $results[] = lang('gateway_management_hosts_not_configured');
            // }

            // setup mybox-redirect
            $configs['dt_mgr']['mybox-redirect'] = 'https://'.$lan_ip.':81';
        }

        return $configs;
    }

    /**
     * Update daemons.
     *
     * @return void
     */
     public function update_daemons()
     {
         clearos_profile(__METHOD__, __LINE__);

         $daemons = array('dnsthingymgr', 'dnsthingyipe');
         if ($this->is_enabled()) {
             foreach ($daemons as $daemon) {
                 // Enable daemon autostart scripts
                 if (!$this->_daemon_cmd($daemon, 'get_boot_state')) {
                     $this->_daemon_cmd($daemon, 'set_boot_state', TRUE);
                     $this->add_issue(lang('gateway_management_'.$daemon.'_boot_not_set'));
                 }

                 // Start the daemons
                 try {
                     $running = $this->_daemon_cmd($daemon, 'get_running_state');
                 } catch (Engine_Exception $e) {
                     // First attempt can raise exception if condition changed recently.
                     $running = $this->_daemon_cmd($daemon, 'get_running_state');
                 }
                 if (!$running) {
                     $this->_daemon_cmd($daemon, 'set_running_state', TRUE);
                     $this->add_issue(lang('gateway_management_'.$daemon.'_not_running'));
                 }
             }
         } else {
             foreach ($daemons as $daemon) {
                 // Disable daemon autostart scripts
                 if (!$this->_daemon_cmd($daemon, 'get_boot_state')) {
                     $this->_daemon_cmd($daemon, 'set_boot_state', FALSE);
                 }

                 // Stop the daemons
                 try {
                     $running = $this->_daemon_cmd($daemon, 'get_running_state');
                 } catch (Engine_Exception $e) {
                     // First attempt can raise exception if condition changed recently.
                     $running = $this->_daemon_cmd($daemon, 'get_running_state');
                 }
                 if ($running) {
                     $this->_daemon_cmd($daemon, 'set_running_state', FALSE);
                 }
             }
         }
     }

    /**
     * Check if network is gateway.
     *
     * @return bool
     */
    public function is_gateway()
    {
        clearos_profile(__METHOD__, __LINE__);

        return $this->is_gateway;
    }

    /**
    * Get DNS IPs.
    *
    * @return array
    */
   public function get_dns()
   {
       clearos_profile(__METHOD__, __LINE__);

       $resolver = new Resolver();
       $resolvers = $resolver->get_nameservers();

       return $resolvers;
   }

   /**
    * Get internal DNS IPs.
    *
    * @return array
    */
   public function get_internal_dns()
   {
       clearos_profile(__METHOD__, __LINE__);

       $internal_dns = array();

       $resolvers = $this->get_dns();
       foreach ($resolvers as $ip) {
           if ($this->is_internal_ip($ip)) {
               $internal_dns[] = $ip;
           }
       }

       return $internal_dns;
   }

   /**
     * Check if an internal IP is used for DNS.
     *
     * @return bool
     */
    public function internal_dns()
    {
        clearos_profile(__METHOD__, __LINE__);

        if ($this->internal_dns === NULL) {
            $this->internal_dns = FALSE;

            $resolvers = $this->get_dns();
            foreach ($resolvers as $ip) {
                if ($this->is_internal_ip($ip))
                    $this->internal_dns = TRUE;
            }
        }

        return $this->internal_dns;
    }

    /**
     * Check if an IP is internal.
     *
     * @return bool
     */
    public function is_internal_ip($ip)
    {
        clearos_profile(__METHOD__, __LINE__);

        $ip_path = $this->get_path('ip');
        $grep_path = $this->get_path('grep');

        // check the ip
        $shell = new Shell();
        $shell->execute($ip_path, "route get $ip | $grep_path -o -P '(?<=dev )[a-zA-Z0-9]*'");

        $result = $shell->get_first_output_line();

        if ($result == 'lo')
            // local interface
            return TRUE;

        $if = $this->get_interface($result);
        if ($if['role'] === 'EXTIF')
            return FALSE;
        else
            return TRUE;
    }

    /**
     * Find the path of a command.
     *
     * @return string path
     */
    public function get_path($cmd)
    {
        clearos_profile(__METHOD__, __LINE__);

        $path = $cmd . '_path';
        if (!$this->$path) {
            $shell = new Shell();
            $shell->execute('/usr/bin/which', "--skip-alias $cmd");
            $this->$path = $shell->get_first_output_line();
        }

        return $this->$path;
    }

    /**
     * Verify firewall is set correctly.
     *
     * @return string error message
     */
    public function update_firewall()
    {
        clearos_profile(__METHOD__, __LINE__);

        $config = $this->config();

        // get the rules
        if ($this->is_enabled()) {
            $fw = $this->_get_firewall_enabled();
        } else {
            $fw = $this->_get_firewall_disabled();
        }

        // make sure file exists
        $file = new File(self::FIREWALL_RULES, TRUE);
        if (!$file->exists()) {
            $file->create('webconfig', 'webconfig', '0755');
        }

        // compare contents
        $current_contents = $file->get_contents();
        if ($current_contents != $fw) {
            $file->delete_lines('/.*/');
            $file->add_lines($fw);
            return lang('gateway_management_firewall_error');
        }
    }

    /**
     * Get firewall rules for when this is enabled.
     *
     * @return string rules
     */
    private function _get_firewall_enabled()
    {
        clearos_profile(__METHOD__, __LINE__);

        $config = $this->config();

        // IPv4 rules (zero-length test for ClearOS 6.6 compatibility)
        $fw = 'if [ "$FW_PROTO" == "ipv4" ] || [ -z "$FW_PROTO" ]; then' . "\n";

        // Set up ipset
        $fw .= "ipset create -exist dt_wa4 hash:ip family ipv4\n";

        // Setup wan access (AKA: No Internet) and IPE (AKA: DTTS)
        $fw .= "  \$IPTABLES -N DNSthingyIPE\n";
        $fw .= "  \$IPTABLES -N DNSthingyEST\n";
        $fw .= "  \$IPTABLES -N DNSthingyREJECT\n";

        // Reject rules
        $fw .= "  \$IPTABLES -A DNSthingyREJECT -p tcp -j REJECT --reject-with tcp-reset\n";
        $fw .= "  \$IPTABLES -A DNSthingyREJECT -p icmp -j REJECT --reject-with icmp-host-unreachable\n";
        $fw .= "  \$IPTABLES -A DNSthingyREJECT -j DROP\n";

        // Handle established connections.
        // Disabling Internet access cuts these off.
        $fw .= "  rule=\$(\$IPTABLES -t filter -L FORWARD -nv --line-numbers | grep -m 1 ESTABLISHED | cut -d' ' -f1)\n";
        $fw .= "  \$IPTABLES -t filter -R FORWARD \$rule -m state --state RELATED,ESTABLISHED -j DNSthingyEST\n";
        $fw .= "  \$IPTABLES -A DNSthingyEST -m set --match-set dt_wa4 src -j DNSthingyREJECT\n";
        $fw .= "  \$IPTABLES -A DNSthingyEST -j ACCEPT\n";

        // Jump to IPE rules (after established connections)
        $fw .= "  ((rule = rule+1))\n";
        $fw .= "  \$IPTABLES -t filter -I FORWARD \$rule -j DNSthingyIPE\n";

        // Block page IP
        $blockpage_ip = $config['http-listen-address'];
        $blockpage_port = self::BLOCKPAGE_PORT;

        $fw .= "  \$IPTABLES -t filter -N DNSthingyBP\n";
        $fw .= "  \$IPTABLES -t filter -I INPUT -d $blockpage_ip -j DNSthingyBP\n";

        $fw .= "  \$IPTABLES -t filter -I DNSthingyBP 1 -p tcp -m multiport --dport 80,6373 -j ACCEPT\n";
        $fw .= "  \$IPTABLES -t filter -I DNSthingyBP 2 -p tcp --dport 443 -j REJECT --reject-with=tcp-reset\n";
        $fw .= "  \$IPTABLES -t filter -I DNSthingyBP 3 -p tcp -j REJECT\n";
        $fw .= "  \$IPTABLES -t filter -I DNSthingyBP 4 -p udp -j REJECT\n";

        // For DNS hijack
        $fw .= "  \$IPTABLES -t nat -N DNSthingy\n";
        $fw .= "  \$IPTABLES -t nat -I PREROUTING -j DNSthingy\n";
        $fw .= "  \$IPTABLES -t nat -I DNSthingy 1 -p tcp -d $blockpage_ip --dport 80 -j DNAT --to $blockpage_ip:$blockpage_port\n";

        // Per-interface rules
        $lan_ifs = $this->get_lan_interfaces();
        foreach ($lan_ifs as $lan_if) {
            $lan_data = $this->get_interface($lan_if);
            $lan_ip = $lan_data['ifcfg']['ipaddr'];

            // Hijack DNS queries
            $fw .= "  \$IPTABLES -t nat -I DNSthingy 1 -p tcp -i $lan_if ! -d $lan_ip --dport 53 -j DNAT --to $lan_ip:53\n";
            $fw .= "  \$IPTABLES -t nat -I DNSthingy 1 -p udp -i $lan_if ! -d $lan_ip --dport 53 -j DNAT --to $lan_ip:53\n";
        }

        // Using a DNS server on the LAN disables hijacking for that IP.
        if ($this->internal_dns()) {
            $internal_ips = $this->get_internal_dns();
            foreach ($internal_ips as $ip) {
                $fw .= "  \$IPTABLES -t nat -I DNSthingy -p tcp --dport 53 -s $ip -j ACCEPT\n";
                $fw .= "  \$IPTABLES -t nat -I DNSthingy -p udp --dport 53 -s $ip -j ACCEPT\n";
            }
        }

        // IPv6 rules (for future use)
        // $fw .= 'elif [ "$FW_PROTO" == "ipv6" ]; then' . "\n";

        // None yet
        $fw .= "fi\n";

        // Make sure we have the required kernel module
        $fw .= "modprobe nfnetlink_queue\n";

        // The firewall rules will sometimes be wiped and reloaded.
        // When this happens, we need to restart IPE server to rebuild IPE rules.
        $fw .= "service dnsthingyipe restart";

        // remove excess new lines
        $fw = trim($fw);

        return $fw;
    }

    /**
     * Get firewall rules for when this is disabled.
     *
     * @return string rules
     */
    private function _get_firewall_disabled()
    {
        clearos_profile(__METHOD__, __LINE__);

        return '';
    }

    /**
     * Get all network interfaces.
     *
     * @return array interfaces
     */
    public function get_all_interfaces()
    {
        clearos_profile(__METHOD__, __LINE__);

        $options['filter_virtual'] = FALSE;
        $iface_manager = new Iface_Manager();
        $interfaces = $iface_manager->get_interface_details($options);

        $this->interfaces = $interfaces;
        return $interfaces;
    }

    /**
     * Get specific network interfaces.
     *
     * @param string $if interface
     *
     * @return array interfaces
     */
    public function get_interface($if)
    {
        clearos_profile(__METHOD__, __LINE__);

        $interfaces = $this->get_all_interfaces();

        if ($interfaces[$if])
            return $interfaces[$if];
        else
            return FALSE;
    }

    /**
     * Get a LAN interface.
     *
     * @return string interface
     */
    public function get_lan_interface()
    {
        $config = $this->config();

        $interfaces = $this->get_lan_interfaces();
        return $interfaces[0];
    }

    /**
     * Get all LAN interfaces.
     *
     * @return array interfaces
     */
    public function get_lan_interfaces()
    {
        $ifs = array();

        $interfaces = $this->get_physical_interfaces();
        foreach ($interfaces as $if => $details) {
            // Only support configured interfaces
            if (!$details['configured'])
                continue;

            // Have to have an IP address
            if (empty($details['address']))
                continue;

            if ($this->is_gateway() && $details['role'] !== 'EXTIF') {
                $ifs[] = $if;
            } elseif (!$this->is_gateway() && $details['role'] === 'EXTIF') {
                $ifs[] = $if;
            }
        }

        return $ifs;
    }

    /**
     * Get physical network interfaces.
     *
     * @return array interfaces
     */
    public function get_physical_interfaces()
    {
        clearos_profile(__METHOD__, __LINE__);

        $interfaces = $this->get_all_interfaces();

        foreach ($interfaces as $if => $details) {
            if ($details['configured'] && $details['type'] !== Iface::TYPE_VIRTUAL)
                $physical[$if] = $details;
        }

        return $physical;
    }

    /**
     * Get external network interfaces.
     *
     * @return array interfaces
     */
    public function get_external_interfaces()
    {
        clearos_profile(__METHOD__, __LINE__);

        $external_ifs = array();

        $interfaces = $this->get_all_interfaces();
        foreach ($interfaces as $if => $details) {
            if ($details['configured'] && $details['role'] === 'EXTIF') {
                $external_ifs[$if] = $details;
            }
        }

        return $external_ifs;
    }

    /**
     * Clear all stored interfaces.
     *
     * @return void
     */
    public function clear_interfaces()
    {
        clearos_profile(__METHOD__, __LINE__);

        unset($this->interfaces);
    }

    /**
     * Returns the mac address in decimal format.
     *
     * @return string mac address in decimal format
     */
    public function get_box_id()
    {
        clearos_profile(__METHOD__, __LINE__);

        $config = $this->config();

        // Later versions store this as a config option
        if (isset($config['boxid']))
            return $config['boxid'];

        $if = $this->get_lan_interface();

        if (!$if)
            return 0;

        // get mac address
        $if = $this->get_interface($if);
        $mac = $if['hwaddress'];

        // make sure mac address was retrieved
        if (!$mac)
            return 0;

        // convert to decimal
        $clean_mac = str_replace(':', '', $mac);
        $box_id = number_format(hexdec($clean_mac), 0, '', '');

        return $box_id;
    }

    /**
     * Returns the ClearOS subscription status.
     *
     * @return string ClearOS subscription status
     */
    public function get_subscription_status()
    {
        clearos_profile(__METHOD__, __LINE__);

        $rest = new Rest();

        $box_id = $this->get_box_id();

        if (!$box_id) {

            $array = array(
                'status' => -1
            );

        } else {

            $extras['boxid'] = $box_id;

            $result = $rest->request('dt_mgr', 'activate', $extras);
            $this->subscription_status = $result;

            $array = json_decode($result, TRUE);
            $array['short_msg'] = lang('gateway_management_sub_status_'.$array['code']);

        }

        $result = json_encode($array);

        echo $result;
        exit;
    }

    /**
     * Restart daemons to reload config.
     *
     * @return void
     */
    public function restart_daemons()
    {
        clearos_profile(__METHOD__, __LINE__);

        $daemons = array('dnsmasq', 'dnsthingymgr', 'dnsthingyipe');

        foreach ($daemons as $daemon) {
            $this->_daemon_cmd($daemon, 'restart');
        }
    }

    /**
     * Suggest an IP for blockpage.
     *
     * @return string IP
     */
    public function suggest_http_address()
    {
        clearos_profile(__METHOD__, __LINE__);

        if ($this->is_gateway()) {
            return $this->_suggest_gateway_http_address();
        } else {
            return $this->_suggest_standalone_http_address();
        }
    }

    /**
     * Suggest an IP for block page in gateway mode.
     *
     * @return string IP
     */
    private function _suggest_gateway_http_address()
    {
        clearos_profile(__METHOD__, __LINE__);

        // use a default
        $ip = '172.27.27.27';
        if (!$this->validate_http_address($ip))
            return $ip;

        // just find anything
        for ($a = 16; $a <= 32; $a++) {
            for ($b = 1; $b <= 255; $b += 50) {
                $ip = "172.$a.$b.172";
                if (!$this->validate_http_address($ip))
                    return $ip;
            }
        }
    }

    /**
     * Suggest an IP for block page in standalone mode.
     *
     * @return string IP
     */
    private function _suggest_standalone_http_address()
    {
        clearos_profile(__METHOD__, __LINE__);

        // get a lan interface
        $lan_if = $this->get_lan_interface();

        // get iface info
        $iface = new Iface($lan_if);
        $live_ip = $iface->get_live_ip();

        // find an acceptable IP
        $part = explode('.', $live_ip);
        for ($i = 1; $i <= 255; $i++) {
            $ip = "$part[0].$part[1].$part[2].$i";
            if (!$this->validate_http_address($ip, $lan_if))
                return $ip;
        }
    }

    /**
     * Validate enabled.
     *
     * @param bool $enabled status
     */
    public function validate_enabled($enabled)
    {
        clearos_profile(__METHOD__, __LINE__);
    }

    /**
     * Validate IP for blockpage.
     *
     * @param string $validate_ip IP address to validate
     *
     * @return string error message if IP is not valid
     */
    public function validate_http_address($validate_ip)
    {
        clearos_profile(__METHOD__, __LINE__);

        // validate IP
        $resolver = new Resolver();
        $result = $resolver->validate_ip($validate_ip);
        if ($result)
            return $result;

        // check that this IP makes sense
        if ($this->is_gateway()) {
            $result = $this->_validate_gateway_http_address($validate_ip);
        } else {
            $result = $this->_validate_standalone_http_address($validate_ip);
        }

        // if validation failed, return that
        if ($result) {
            return $result;
        }

        // if this is not set, make sure it is available.
        if ($validate_ip !== $this->get_http_address()) {
            if ($this->is_ip_in_use($validate_ip)) {
                return lang('gateway_management_blockpage_ip_in_use');
            }
        }
    }

    /**
     * Validate IP for gateway block page.
     *
     * @param string $validate_ip IP address to validate
     *
     * @return string error message if IP is not valid
     */
    private function _validate_gateway_http_address($validate_ip, $check_use=TRUE)
    {
        clearos_profile(__METHOD__, __LINE__);

        $config = $this->config();

        // validate IP is not on any networks
        $result = $this->ip_on_networks($validate_ip);
        if ($result) {
            // can use a local IP address if conditions are met
            $result = $this->_validate_standalone_http_address($validate_ip);
            return $result;
        }

        // Make sure this is a private IP address
        $network_utils = new Network_Utils();
        if (
            !$network_utils->ip_on_network($validate_ip, '10.0.0.0/8') &&
            !$network_utils->ip_on_network($validate_ip, '172.16.0.0/12') &&
            !$network_utils->ip_on_network($validate_ip, '192.168.0.0/16')
        ) {
            return lang('gateway_management_blockpage_ip_not_private');
        }
    }

    /**
     * Validate IP for standalone block page.
     *
     * @param string $validate_ip IP address to validate
     *
     * @return string error message if IP is not valid
     */
    private function _validate_standalone_http_address($validate_ip)
    {
        clearos_profile(__METHOD__, __LINE__);

        $config = $this->config();

        $lans = $this->get_lan_interfaces();

        foreach ($lans as $lan_if) {
            // validate IP is on subnet
            $iface = new Iface($lan_if);
            $network = $iface->get_network();

            // If we don't have network info, skip this
            if (empty($network)) {
                continue;
            }

            // See if the IP belongs to this network
            $network_utils = new Network_Utils();
            if (!$network_utils->ip_on_network($validate_ip, $network)) {
                continue;
            }

            // check that IP is outside DHCP range
            $dnsmasq = new Dnsmasq();
            $range = $dnsmasq->get_subnet($lan_if);
            $ip_long = ip2long($validate_ip);
            if ($ip_long >= ip2long($range['start']) && $ip_long <= ip2long($range['end'])) {
                return lang('dnsthingy_blockpage_ip_in_dhcp_range')." ($range[start] - $range[end])";
            }

            // IP is valid
            return;
        }

        // IP wasn't on any of the LAN networks
        return lang('dnsthingy_blockpage_ip_wrong_subnet');
    }

    /**
     * Check if the given IP is on any networks.
     *
     * @param string $validate_ip IP address to validate
     *
     * @return bool true if on a network
     */
    public function ip_on_networks($validate_ip)
    {
        clearos_profile(__METHOD__, __LINE__);

        $config = $this->config();

        // validate IP is not on any networks
        $ifaces = $this->get_all_interfaces();
        foreach ($ifaces as $iface => $info) {
            // skip any inactive interfaces
            if (!$info['address'])
                continue;

            $iface = new Iface($iface);
            $network = $iface->get_network();

            // external interfaces have an empty string as network
            if (empty($network))
                continue;

            $network_utils = new Network_Utils();
            if ($network_utils->ip_on_network($validate_ip, $network))
                return TRUE;
        }
    }

    /**
     * Make sure an IP address is available.
     *
     * @param string $ip IP address to check
     *
     * @return bool
     */
    public function is_ip_in_use($ip)
    {
        clearos_profile(__METHOD__, __LINE__);

        $shell = new Shell();

        // find ping
        if (!$this->ping_path) {
            try {
                $shell->execute('/usr/bin/which', 'ping');
                $this->ping_path = $shell->get_first_output_line();
            } catch (Engine_Exception $e) {
                // clearos 7 default
                $this->ping_path = '/usr/bin/ping';
            }
        }

        // initialize status
        $status = TRUE;

        // ping the IP
        try {
            // 0 means response received, anything else is failure.
            $exit_code = $shell->execute($this->ping_path, "-c 1 $ip");
            $status = !$exit_code;
        } catch (Engine_Exception $e) {
            // no response, so probably not in use
            $status = FALSE;
        }

        // For unresponsive internal IP addresses, check ARP as well.
        if (!$status && $this->is_internal_ip($ip)) {
            try {
                $shell->execute('/sbin/arp', "-a -n | grep '$ip) at <incomplete>'");
                $status = FALSE;
            } catch (Engine_Exception $e) {
                // not critical
            }
        }

        return $status;
    }

    /**
     * Returns the config.
     *he
     * @param string $which Which config file
     *
     * @return array config file
     */

    public function config($which='dt_mgr')
    {
        clearos_profile(__METHOD__, __LINE__);

        if (empty($this->is_loaded[$which])) {
            $this->_load_config($which);
        }
        return $this->config[$which];
    }


    ///////////////////////////////////////////////////////////////////////////////
    // P R I V A T E   M E T H O D S
    ///////////////////////////////////////////////////////////////////////////////

    /**
     * Create all the necessary keys.
     *
     * @return void
     */
    protected function _create_keys() {
        $shell = new Shell();

    	if (!file_exists(self::PATH_PK_FILE) || !file_exists(self::PATH_SK_FILE)) {
    		echo "Creating key pair.\n";
            $shell->execute('/usr/local/sbin/dnsthingyckg', self::PATH_PK_FILE . ' ' . self::PATH_SK_FILE);
    	}

    	if (!file_exists(self::PEM_CERT_FILE) || !file_exists(self::PEM_KEY_FILE) || !file_exists(self::PEM_DH_FILE)) {
    		echo "Creating Diffie–Hellman key. This can take a few minutes on slower hardware.\n";
            $shell->execute('/usr/bin/openssl', 'dhparam -out ' . self::PEM_DH_FILE . ' 1024');
    		echo "Creating certificates.\n";

    		$dn = array(
    			'countryName' => 'CA',
    			'stateOrProvinceName' => 'ON',
    			'localityName' => 'London',
    			'organizationName' => 'DNSthingy Inc.',
    			'organizationalUnitName' => 'Firmware',
    			'commonName' => 'services',
    			'emailAddress' => 'service@dnsthingy'
    		);
    		$privkey = openssl_pkey_new();
    		$csr = openssl_csr_new($dn, $privkey);
    		$cert = openssl_csr_sign($csr, NULL, $privkey, 2555);
    		openssl_x509_export_to_file($cert, self::PEM_CERT_FILE);
    		openssl_pkey_export_to_file($privkey, self::PEM_KEY_FILE, 'dnsthingyIPE');
    	}

    	/* Ensure that the self-signed cert is listed in cert.pem */
        $ca_cert = file_get_contents('/etc/dnsthingy/cert.pem');
        $pem_cert = file_get_contents(self::PEM_CERT_FILE);
        if (strpos($ca_cert, $pem_cert) === FALSE) {
        	echo "Adding to chain.\n";
        	file_put_contents('/etc/dnsthingy/cert.pem', $pem_cert, FILE_APPEND);
        }
    }

    /**
     * Setup the DNSthingy config file
     *
     * @return void
     */
    protected function _setup_dt_config()
    {
        clearos_profile(__METHOD__, __LINE__);

        $config = $this->config();

        // CLEANUP old settings
        // remove listen-iface
        if (isset($config['listen-iface'])) {
            $this->_set_parameter('dt_mgr', 'listen-iface', NULL, TRUE);
        }

        // remove listen-address
        if (isset($config['listen-address'])) {
            $this->_set_parameter('dt_mgr', 'listen-address', NULL, TRUE);
        }

        // remove old hosts entries
        $hosts = new Hosts();
        $host_ip = $hosts->get_ip_by_hostname('test.dnsthingy.com');
        if ($host_ip) {
            $hosts->delete_entry($host_ip);
        }

        // Setup new settings
        // set box_id
        if (!isset($config['boxid'])) {
            $box_id = $this->get_box_id();
            if ($box_id > 0) {
                $this->_set_parameter('dt_mgr', 'boxid', $box_id);
            } else {
                $this->issues[] = 'Unable to find LAN interface.';
                return FALSE;
            }
        }

        // move redirect-local
        if (isset($config['redirect-local'])) {
            $this->_set_parameter('dt_mgr', 'http-listen-address', $config['redirect-local']);
            $this->_set_parameter('dt_mgr', 'redirect-local', NULL, TRUE);
        }
    }

    /**
     * Setup the IPE config file
     *
     * @return void
     */
    protected function _setup_ipe_config()
    {
        clearos_profile(__METHOD__, __LINE__);

        $config = $this->config();

    }

    /**
     * return proper config file.
     *
     * @param string $which Which config file
     *
     * @return  string config file location
     * @throws Engine_Exception
     */
    protected function _select_config($which)
    {
        clearos_profile(__METHOD__, __LINE__);

        switch ($which) {
            case 'dt_mgr':
                return self::DT_MGR_CONFIG;
            case 'dt_ipe':
                return self::DT_IPE_CONFIG;
            case 'dt_app':
                return self::DT_APP;
            case 'dnsmasq':
                return self::DNSMASQ_CONFIG;
            case 'dnsmasq_main':
                return self::DNSMASQ_MAIN_CONFIG;
            default:
                throw new Engine_Exception("Invalid config type: '$which'.", CLEAROS_ERROR);
        }
    }

    /**
     * Loads configuration files.
     *
     * @param string $which Which config file
     *
     * @return  void
     * @throws Engine_Exception
     */
    protected function _load_config($which)
    {
        clearos_profile(__METHOD__, __LINE__);

        $location = $this->_select_config($which);

        try {
            $folder = new Folder(self::DNSTHINGY_DIR);
            if (!$folder->exists())
                $folder->create('root', 'root', '0755');

            $file = new Configuration_File($location, 'match', "/([^=]*)\s*=?\s*(.*)?/");

            if (!$file->exists())
                $file->create('webconfig', 'webconfig', '0644');

            $this->config[$which] = $file->load();
        } catch (Exception $e) {
            throw new Engine_Exception(clearos_exception_message($e), CLEAROS_ERROR);
        }

        $this->is_loaded[$which] = TRUE;
    }

    /**
     * Setup a hosts entry.
     *
     * @return bool
     */
    protected function _setup_host($host, $ip, $aliases = array())
    {
        $hosts = new Hosts();

        // remove invalid IPs for these hosts
        $all_hosts = array_merge(array($host), $aliases);
        foreach ($all_hosts as $current_host) {
            while (
                $hosts->get_ip_by_hostname($current_host)
                && $hosts->get_ip_by_hostname($current_host) != $ip
            ) {
                $hosts->delete_entry($hosts->get_ip_by_hostname($current_host));
            }
        }

        // setup host (if needed)
        foreach ($all_hosts as $current_host) {
            if (!$hosts->get_ip_by_hostname($current_host)) {
                if ($hosts->entry_exists($ip)) {
                    $hosts->edit_entry($ip, $host, $aliases);
                } else {
                    $hosts->add_entry($ip, $host, $aliases);
                }
                return TRUE;
            }
        }

        return FALSE;
    }

    /**
     * Verify a parameter is correct.
     *
     * @param string $which file
     * @param string $key   key name
     * @param string $value value for the key
     * @param string $delete erase the key (if it exists)
     *
     * @return bool
     * @throws Engine_Exception
     */
    protected function _verify_parameter($which, $key, $value=NULL, $delete=FALSE)
    {
        clearos_profile(__METHOD__, __LINE__);

        $config = $this->config($which);

        if (isset($config[$key])) {
            // remove comments, must be ' #' to avoid removing dnsmasq port numbers
            $old_values = explode(' #', $config[$key]);
            $old_value = $old_values[0];
        } else {
            $old_value = NULL;
        }

        // check if values unchanged
        // non-null
        if ($value !== NULL && $old_value === $value)
            return TRUE;
        // null
        elseif ($value === NULL && $old_value === '')
            return TRUE;
        // delete
        elseif (!isset($config[$key]) && $delete)
            return TRUE;

        // change value, storing old value if needed
        if ($which == 'dnsmasq_main' && isset($old_value)) {
            if ($delete) {
                $this->_set_parameter($which, $key, "#DNSthingy disabled:$key", $delete);
            } else {
                $this->_set_parameter($which, $key, "$value #DNSthingy changed from:$old_value", $delete);
            }
        } else {
            $this->_set_parameter($which, $key, $value, $delete);
        }
        return FALSE;
    }


    /**
     * Reset a parameter to pre-install state.
     *
     * @param string $which file
     * @param string $key   key name
     *
     * @return bool
     * @throws Engine_Exception
     */
    protected function _reset_parameter($which, $key)
    {
        clearos_profile(__METHOD__, __LINE__);

        $config = $this->config($which);

        if (isset($config[$key])) {
            // key found by DNSthingy
            $old_value = $config[$key];

            if (strpos($old_value, ' #DNSthingy changed from:') === FALSE) {
                // key wasn't changed by DNSthingy
                return FALSE;
            }

            // remove current value, must be ' #' to avoid removing dnsmasq port numbers
            $old_value = explode(' #', $old_value);
            $old_value = $old_value[1];

            // change the setting
            $value = str_replace('DNSthingy changed from:', '', $old_value);
            $this->_set_parameter($which, $key, $value);

            return TRUE;
        }

        // See if a line like this exists. Config strips comment line, so use File.
        $config = $this->_select_config($which);
        $file = new File($config, TRUE);

        // Should never happen, but who knows.
        if (!$file->exists()) {
            return;
        }

        // Clear the value
        $matched = $file->delete_lines("/^#DNSthingy disabled:.*$/", $key);
        if ($matched) {
            $this->_set_parameter($which, $key, NULL);
            return TRUE;
        }

        return FALSE;
    }

    /**
     * Generic set routine.
     *
     * @param string $which  file
     * @param string $key    key name
     * @param string $value  value for the key
     * @param string $delete erase the key
     *
     * @return  void
     * @throws Engine_Exception
     */
    protected function _set_parameter($which, $key, $value=NULL, $delete=FALSE)
    {
        clearos_profile(__METHOD__, __LINE__);

        $location = $this->_select_config($which);

        try {
            $file = new File($location, TRUE);

            if (!$file->exists())
                $file->create('webconfig', 'webconfig', '0644');

            if ($delete) {
                if ($value) {
                    // can be used to comment out lines
                    $file->replace_lines("/^$key\s*(=\s*)?/", "$value\n");
                } else {
                    $file->delete_lines("/^$key\s*(=\s*)?/");
                }
            } else {
                if ($value) {
                    $line = "$key=$value\n";
                } else {
                    $line = "$key\n";
                }

                $matched = $file->replace_lines("/^$key\s*(=\s*)?/", $line);
                if (!$matched) {
                    $file->add_lines($line);
                }
            }

        } catch (Exception $e) {
            throw new Engine_Exception(clearos_exception_message($e), CLEAROS_ERROR);
        }

        $this->is_loaded[$which] = FALSE;
    }


    /**
     * Runs the specified daemon command.
     *
     * @param string $script  Name
     * @param string $command Command
     * @param string $value Value (only needed for some items)
     *
     * @return string/bool
     * @throws Engine_Exception
     */
    protected function _daemon_cmd($script, $command, $value=FALSE)
    {
        clearos_profile(__METHOD__, __LINE__);

        $daemon = new Daemon($script);

        switch ($command) {
            case 'get_boot_state':
                return $daemon->get_boot_state();
            case 'set_boot_state':
                return $daemon->set_boot_state($value);
            case 'restart':
                return $daemon->restart();
            case 'get_status':
                return $daemon->get_status(); // text
            case 'get_running_state':
                return $daemon->get_running_state(); // code
            case 'set_running_state':
                return $daemon->set_running_state($value);
            default:
                throw new Engine_Exception("Invalid command: '$command'.", CLEAROS_ERROR);
        }
    }

}
